--steyer8300 
--hz888


steyer8300 = {};

function steyer8300.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Motorized, specializations) and SpecializationUtil.hasSpecialization(Hirable, specializations);
end;

function steyer8300:load(xmlFile)

    local fulkenode = Utils.indexToObject(self.rootNode, getXMLString(xmlFile, "vehicle.fulke#index"));
    if fulkenode ~= nil then
        self.fulke = {};
        self.fulke.node = fulkenode;

        local x, y, z = Utils.getVectorFromString(getXMLString(xmlFile, "vehicle.fulke#minRot"));
        self.fulke.minRot = {};
        self.fulke.minRot[1] = Utils.degToRad(Utils.getNoNil(x, 0));
        self.fulke.minRot[2] = Utils.degToRad(Utils.getNoNil(y, 0));
        self.fulke.minRot[3] = Utils.degToRad(Utils.getNoNil(z, 0));

        x, y, z = Utils.getVectorFromString(getXMLString(xmlFile, "vehicle.fulke#maxRot"));
        self.fulke.maxRot = {};
        self.fulke.maxRot[1] = Utils.degToRad(Utils.getNoNil(x, 0));
        self.fulke.maxRot[2] = Utils.degToRad(Utils.getNoNil(y, 0));
        self.fulke.maxRot[3] = Utils.degToRad(Utils.getNoNil(z, 0));

        self.fulke.rotTime = Utils.getNoNil(getXMLString(xmlFile, "vehicle.fulke#rotTime"), 2)*1000;
        self.fulke.touchRotLimit = Utils.degToRad(Utils.getNoNil(getXMLString(xmlFile, "vehicle.fulke#touchRotLimit"), 10));
    end;

    local ajtonode = Utils.indexToObject(self.rootNode, getXMLString(xmlFile, "vehicle.ajto#index"));
    if ajtonode ~= nil then
        self.ajto = {};
        self.ajto.node = ajtonode;

        local x, y, z = Utils.getVectorFromString(getXMLString(xmlFile, "vehicle.ajto#minRot"));
        self.ajto.minRot = {};
        self.ajto.minRot[1] = Utils.degToRad(Utils.getNoNil(x, 0));
        self.ajto.minRot[2] = Utils.degToRad(Utils.getNoNil(y, 0));
        self.ajto.minRot[3] = Utils.degToRad(Utils.getNoNil(z, 0));

        x, y, z = Utils.getVectorFromString(getXMLString(xmlFile, "vehicle.ajto#maxRot"));
        self.ajto.maxRot = {};
        self.ajto.maxRot[1] = Utils.degToRad(Utils.getNoNil(x, 0));
        self.ajto.maxRot[2] = Utils.degToRad(Utils.getNoNil(y, 0));
        self.ajto.maxRot[3] = Utils.degToRad(Utils.getNoNil(z, 0));

        self.ajto.rotTime = Utils.getNoNil(getXMLString(xmlFile, "vehicle.ajto#rotTime"), 2)*1000;
        self.ajto.touchRotLimit = Utils.degToRad(Utils.getNoNil(getXMLString(xmlFile, "vehicle.ajto#touchRotLimit"), 10));
    end;
	
    self.controlpath = Utils.getFilename("control.png", self.baseDirectory);
    self.controlhudPoxX = 0.01;
    self.controlhudPoxY = 1-0.204;
    self.controlhudWidth = 0.270;
    self.controlhudHeight = 0.270;
    self.controlhudPoxX = 0.48;
    self.controlhudPoxY = 0.8
    self.controlhudOverlay = Overlay:new("hudControl", self.controlpath, self.controlhudPoxX, self.controlhudPoxY, self.controlhudWidth, self.controlhudHeight);
    self.panelActive = false;
    self.fulkefor = false;

--ai

    self.startAITractor = SpecializationUtil.callSpecializationsFunction("startAITractor");
    self.stopAITractor = SpecializationUtil.callSpecializationsFunction("stopAITractor");
    self.onTrafficCollisionTrigger = steyer8300.onTrafficCollisionTrigger;
    self.onToolTrafficCollisionTrigger = steyer8300.onToolTrafficCollisionTrigger;

    self.isAITractorActivated = false;
    self.aiTractorDirectionNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.aiTractorDirectionNode#index"));
    if self.aiTractorDirectionNode == nil then
        self.aiTractorDirectionNode = self.components[1].node;
    end;

    self.aiTractorLookAheadDistance = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.aiTractorLookAheadDistance"), 10);
    self.turnTimeout = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.turnTimeout"), 800);
    self.turnTimeoutLong = self.turnTimeout*10;
    self.turnTimer = self.turnTimeoutLong;
    self.frontMarkerDistanceScale = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.frontMarkerDistanceScale"), 1.1);
    self.lastFrontMarkerDistance = 0;
    self.turnTargetMoveBack = 7;
    self.turnEndDistance = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.turnEndDistance"), 4);
    self.turnEndBackDistance = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.turnEndBackDistance"), 0)+self.turnTargetMoveBack; ---self.frontMarkerDistance;

    self.aiTractorTurnRadius = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.aiTractorTurnRadius"), 5);
    self.aiTurnNoBackward = false;


    self.waitForTurnTimeout = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.waitForTurnTime"), 1200);
    self.waitForTurnTime = 0;


    self.turnStage2Timeout = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.turnForwardTimeout"), 20000);
    self.turnStage3Timeout = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.turnBackwardTimeout"), 20000);
    self.turnStage6Timeout = 3000;


    self.sideWatchDirOffset = -8;
    self.sideWatchDirSize = 7;

    self.turnStage = 0;



    self.aiTrafficCollisionTrigger = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.aiTrafficCollisionTrigger#index"));

    self.aiTurnWidthScale = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.aiTurnWidthScale#value"), 0.9);


    self.trafficCollisionIgnoreList = {};
    for k,v in pairs(self.components) do
        self.trafficCollisionIgnoreList[v.node] = true;
    end;
    self.numCollidingVehicles = 0;
    self.numToolsCollidingVehicles = {};

    self.aiToolsDirty = true;

    self.dtSum = 0;

    local aiMotorSound  = getXMLString(xmlFile, "vehicle.aiMotorSound#file");
    if aiMotorSound  ~= nil and aiMotorSound  ~= "" then
        aiMotorSound  = Utils.getFilename(aiMotorSound, self.baseDirectory);
        self.aiMotorSoundPitchOffset = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.motorSoundRun#pitchOffset"), 0);
        self.aiMotorSoundRadius = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.motorSoundRun#radius"), 50);
        self.aiMotorSoundInnerRadius = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.motorSoundRun#innerRadius"), 10);
        self.aiMotorSoundVolume = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.motorSoundRun#volume"), 1);
        self.aiMotorSound = createAudioSource("aiMotorSound", aiMotorSound, self.aiMotorSoundRadius, self.aiMotorSoundInnerRadius, self.aiMotorSoundVolume, 0);
        link(self.components[1].node, self.aiMotorSound);
        setVisibility(self.aiMotorSound, false);
    end;

--steerable

    self.onEnter = SpecializationUtil.callSpecializationsFunction("onEnter");
    self.onLeave = SpecializationUtil.callSpecializationsFunction("onLeave");
    self.drawGrainLevel = SpecializationUtil.callSpecializationsFunction("drawGrainLevel");


    self.enterReferenceNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.enterReferenceNode#index"));
    self.exitPoint = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.exitPoint#index"));


    self.steering = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.steering#index"));
    if self.steering ~= nil then
        self.steeringSpeed = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.steering#rotationSpeed"), 0);
    end;

    self.numCameras = Utils.getNoNil(getXMLInt(xmlFile, "vehicle.cameras#count"), 0);
    if self.numCameras == 0 then
        print("Error: no cameras in xml file: ", configFile);
    end;
    self.cameras = {};
    for i=1, self.numCameras do
        local cameranamei = string.format("vehicle.cameras.camera%d", i);
        local camIndexStr = getXMLString(xmlFile, cameranamei .. "#index");
        local cameraNode = Utils.indexToObject(self.components, camIndexStr);
        local rotatable = getXMLBool(xmlFile, cameranamei .. "#rotatable");
        local limit = getXMLBool(xmlFile, cameranamei .. "#limit");
        local rotMinX = getXMLFloat(xmlFile, cameranamei .. "#rotMinX");
        local rotMaxX = getXMLFloat(xmlFile, cameranamei .. "#rotMaxX");
        local transMin = getXMLFloat(xmlFile, cameranamei .. "#transMin");
        local transMax = getXMLFloat(xmlFile, cameranamei .. "#transMax");
        local rotateNode = "";
        if rotatable then
            rotateNode = Utils.indexToObject(self.components, getXMLString(xmlFile, cameranamei .. "#rotateNode"));
        end;
        self.cameras[i] = VehicleCamera:new(cameraNode, rotatable, rotateNode, limit, rotMinX, rotMaxX, transMin, transMax);
    end;

    self.camIndex = 1;

    self.tipCamera = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.tipCamera#index"));
    self.backcamera = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.backcamera#index"));
    self.characterNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.characterNode#index"));
    if self.characterNode ~= nil then
        self.characterCameraMinDistance = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.characterNode#cameraMinDistance"), 1.5);
        setVisibility(self.characterNode, false);
    end;

    self.speedRotScale = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.speedRotScale#scale"), 80);
    self.speedRotScaleOffset = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.speedRotScale#offset"), 0.7);


    self.isEntered = false;

    self.steeringEnabled = true;
    self.stopMotorOnLeave = true;
    self.disableCharacterOnLeave = true;
    self.deactivateOnLeave = true;
    self.stopRefuelOnLeave = true;
    self.deactivateLightsOnLeave = true;

    self.showWaterWarning = false;
    self.waterSplashSample = nil;

    self.hudBasePoxX = 0.8325;
    self.hudBasePoxY = (1-0.99);
    self.hudBaseWidth = 0.16;
    self.hudBaseHeight = 0.1625;
    self.hudBaseOverlay = Overlay:new("hudBaseOverlay", "dataS/missions/hud_vehicle_base".. g_languageSuffix .. ".png", self.hudBasePoxX, self.hudBasePoxY, self.hudBaseWidth, self.hudBaseHeight);

end;

function steyer8300:delete()

    self:stopAITractor();

    for i=1, table.getn(self.cameras) do
        self.cameras[i]:delete();
    end;

    if self.waterSplashSample ~= nil then
        delete(self.waterSplashSample);
    end;

    if self.hudBaseOverlay then
        self.hudBaseOverlay:delete();
    end;

end;

function steyer8300:mouseEvent(posX, posY, isDown, isUp, button)

    self.cameras[self.camIndex]:mouseEvent(posX, posY, isDown, isUp, button);

end;


function steyer8300:keyEvent(unicode, sym, modifier, isDown)

    if isDown and sym == Input.KEY_4 then 
	self.fulkefor = not self.fulkefor;
    end;

    if isDown and sym == Input.KEY_5 then
	self.ajtonyit = not self.ajtonyit;
    end;

    if isDown and sym == Input.KEY_0 then
	self.panelActive = not self.panelActive;
    end;

end;

function steyer8300:update(dt)

--ai

    if self:getIsActiveForInput() then
        if not g_currentMission.disableTractorAI and g_currentMission.allowSteerableMoving then
            if InputBinding.hasEvent(InputBinding.TOGGLE_AI) then
                if self.isAITractorActivated then
                    self:stopAITractor();
                else
                    self:startAITractor();
                end;
            end;
        end;
    end;

    if self.aiToolsDirty then
        steyer8300.updateToolsInfo(self);
    end;

    if self.isAITractorActivated then
        if self.isBroken then
            self:stopAITractor();
        end;
        self.dtSum = self.dtSum + dt;
        if self.dtSum > 20 then
           steyer8300.updateAIMovement(self, self.dtSum);
            self.dtSum = 0;
        end;
        steyer8300.updateAIMovement(self, dt);
    else
        self.dtSum = 0;
    end;

--steerable

    if self:getIsActive() then
        if self.steering ~= nil then
            setRotation(self.steering, 0, self.rotatedTime*self.steeringSpeed, 0);
        end;

        local xt,yt,zt = getTranslation(self.components[1].node);
        local deltaWater = yt-g_currentMission.waterY+2.5;
        if deltaWater < 0 then
            self.isBroken = true;
            g_currentMission:onSunkVehicle();

            if self.isEntered then
                g_currentMission:onLeaveVehicle();

                if self:getIsActiveForSound() then
                    local volume = math.min(1, self.lastSpeed*3600/30);
                    if self.waterSplashSample == nil then
                        self.waterSplashSample = createSample("waterSplashSample");
                        loadSample(self.waterSplashSample, "data/maps/sounds/waterSplash.wav", false);
                    end;
                    playSample(self.waterSplashSample, 1, volume, 0);
                end;
            end;
        end;
        self.showWaterWarning = deltaWater < 2;

    end;


    if self.isEntered then

        if InputBinding.hasEvent(InputBinding.REFUEL) then
            if self.doRefuel then
                self:stopRefuel();
            else
                self:startRefuel();
            end;
        end;

        if self.characterNode ~= nil then
            local cx, cy, cz = getWorldTranslation(self.characterNode);
            local x,y,z = getWorldTranslation(getCamera());
            local dist = Utils.vector3Length(cx-x, cy-y, cz-z);
            if dist < self.characterCameraMinDistance then
                setVisibility(self.characterNode, false);
            else
                setVisibility(self.characterNode, true);
            end;
        end;

        if not g_currentMission.fixedCamera then
            setCamera(self.cameras[self.camIndex].cameraNode);
            self.cameras[self.camIndex]:update(dt);
        else
            if self.tipCamera ~= nil then
                setCamera(self.tipCamera);
            else
                self.cameras[self.camIndex]:resetCamera();
            end;
        end;

        if self.steeringEnabled then

            local fuelUsed = self.lastMovedDistance*self.fuelUsage;
            self:setFuelFillLevel(self.fuelFillLevel-fuelUsed);

            g_currentMission.missionStats.fuelUsageTotal = g_currentMission.missionStats.fuelUsageTotal + fuelUsed;
            g_currentMission.missionStats.fuelUsageSession = g_currentMission.missionStats.fuelUsageSession + fuelUsed;

            g_currentMission.missionStats.traveledDistanceTotal = g_currentMission.missionStats.traveledDistanceTotal + self.lastMovedDistance*0.001;
            g_currentMission.missionStats.traveledDistanceSession = g_currentMission.missionStats.traveledDistanceSession + self.lastMovedDistance*0.001;

        
            local acceleration = 0;
            if g_currentMission.allowSteerableMoving and not self.playMotorSound then
		if self.fulkefor then
			acceleration = InputBinding.getAnalogInputAxis(InputBinding.AXIS_FORWARD);
                	if InputBinding.isAxisZero(acceleration) then
                    		acceleration = InputBinding.getDigitalInputAxis(InputBinding.AXIS_FORWARD);
                	end;
                	if math.abs(acceleration) > 0.8 then
                    		self.motor:setSpeedLevel(0, true)
                	end;
                	if self.motor.speedLevel ~= 0 then
                    		acceleration = -1.0;
                	end;
			for k, wheel in pairs(self.wheels) do
				wheel.rotSpeed = Utils.degToRad(-45);
			end;
		else
			acceleration = -InputBinding.getAnalogInputAxis(InputBinding.AXIS_FORWARD);
                	if InputBinding.isAxisZero(acceleration) then
                    		acceleration = -InputBinding.getDigitalInputAxis(InputBinding.AXIS_FORWARD);
                	end;
                	if math.abs(acceleration) > 0.8 then
                    		self.motor:setSpeedLevel(0, true)
                	end;
                	if self.motor.speedLevel ~= 0 then
                    		acceleration = 1.0;
                	end;
			for k, wheel in pairs(self.wheels) do
				wheel.rotSpeed = Utils.degToRad(45);
			end;
		end;
            end;
            if self.fuelFillLevel == 0 then
                acceleration = 0;
            end;

            local inputAxisX = InputBinding.getAnalogInputAxis(InputBinding.AXIS_SIDE);
            if not InputBinding.isAxisZero(inputAxisX) then
                if inputAxisX < 0 then
                    self.rotatedTime = math.min(-self.maxRotTime*inputAxisX, self.maxRotTime);
                else
                    self.rotatedTime = math.max(self.minRotTime*inputAxisX, self.minRotTime);
                end;
            else
                local rotScale = math.min(1.0/(self.lastSpeed*self.speedRotScale+self.speedRotScaleOffset), 1);
                local inputAxisX = InputBinding.getDigitalInputAxis(InputBinding.AXIS_SIDE);
                if inputAxisX < 0 then
                    self.rotatedTime = math.min(self.rotatedTime - dt/1000*inputAxisX*rotScale, self.maxRotTime);
                elseif inputAxisX > 0 then
                    self.rotatedTime = math.max(self.rotatedTime - dt/1000*inputAxisX*rotScale, self.minRotTime);
                else
                    if self.autoRotateBackSpeed ~= 0 then
                        if self.rotatedTime > 0 then
                            self.rotatedTime = math.max(self.rotatedTime - dt/1000*self.autoRotateBackSpeed*rotScale, 0);
                        else
                            self.rotatedTime = math.min(self.rotatedTime + dt/1000*self.autoRotateBackSpeed*rotScale, 0);
                        end;
                    end;
                end;
            end;

            if self.firstTimeRun then
                WheelsUtil.updateWheelsPhysics(self, dt, self.lastSpeed, acceleration, false, self.requiredDriveMode)
            end;
        end;

        if InputBinding.hasEvent(InputBinding.SWITCH_IMPLEMENT) then
            local selected = self.selectedImplement;
            local numImplements = table.getn(self.attachedImplements);
            if selected ~= 0 and numImplements > 1 then
                selected = selected+1;
                if selected > numImplements then
                    selected = 1;
                end;
                self:setSelectedImplement(selected);
            end;
        end;

        if not g_currentMission.fixedCamera then
            if InputBinding.hasEvent(InputBinding.CAMERA_SWITCH) then
                self.cameras[self.camIndex]:onDeactivate();
                self.camIndex = self.camIndex + 1;
                if self.camIndex > self.numCameras then
                    self.camIndex = 1;
                end;
                self.cameras[self.camIndex]:onActivate();
            end;
        end;

        if Input.isKeyPressed(Input.KEY_shift) and self.camIndex == 2 then
                setCamera(self.backcamera);
       	end;

        if InputBinding.hasEvent(InputBinding.TOGGLE_LIGHTS) then
            self:setLightsVisibility(not self.lightsActive);
        end;

        if InputBinding.hasEvent(InputBinding.SPEED_LEVEL1) then
            self.motor:setSpeedLevel(1, false);
        elseif InputBinding.hasEvent(InputBinding.SPEED_LEVEL2) then
            self.motor:setSpeedLevel(2, false);
        elseif InputBinding.hasEvent(InputBinding.SPEED_LEVEL3) then
            self.motor:setSpeedLevel(3, false);
        end;

        if InputBinding.hasEvent(InputBinding.ATTACH) then
            self:handleAttachEvent();
        end;

        if InputBinding.hasEvent(InputBinding.LOWER_IMPLEMENT) then
            self:handleLowerImplementEvent();
        end;

	if self.fulke ~= nil then
		local x, y, z = getRotation(self.fulke.node);
		local rot = {x,y,z};
		local newRot = Utils.getMovedLimitedValues(rot, self.fulke.maxRot, self.fulke.minRot, 3, self.fulke.rotTime, dt, not self.fulkefor);
		setRotation(self.fulke.node, unpack(newRot));
	end;
	if self.ajto ~= nil then
		local x, y, z = getRotation(self.ajto.node);
		local rot = {x,y,z};
		local newRot = Utils.getMovedLimitedValues(rot, self.ajto.maxRot, self.ajto.minRot, 3, self.ajto.rotTime, dt, not self.ajtonyit);
		setRotation(self.ajto.node, unpack(newRot));
	end;
    end;
 
end;

function steyer8300:draw()

--ai

    if not g_currentMission.disableTractorAI then
        if self.aiLeftMarker ~= nil and self.aiRightMarker ~= nil and self.aiBackMarker ~= nil then
            if self.isAITractorActivated then
                g_currentMission:addHelpButtonText(g_i18n:getText("DismissEmployee"), InputBinding.TOGGLE_AI);
            else
                g_currentMission:addHelpButtonText(g_i18n:getText("HireEmployee"), InputBinding.TOGGLE_AI);
            end;
        end;
    end;

--steerable

    local kmh = self.lastSpeed*self.speedDisplayScale*3600;

    self.hudBaseOverlay:render();

    setTextBold(true);

    setTextColor(1.0, 1.0, 1.0, 1.0);
    if kmh > 0 and kmh < 100 then
        renderText(self.hudBasePoxX+0.007, self.hudBasePoxY+0.095, 0.06, string.format("%2d", kmh));
    end;
    renderText(self.hudBasePoxX+0.058, self.hudBasePoxY+0.094, 0.06, g_i18n:getText("speedometer"));
    renderText(self.hudBasePoxX+0.031, self.hudBasePoxY+0.071, 0.03, string.format("%d", g_currentMission.missionStats.money));

    local fuelWarn = 50;
    if self.fuelFillLevel < fuelWarn then
        setTextColor(1,0,0,1);
    end;
    renderText(self.hudBasePoxX+0.031, self.hudBasePoxY+0.039, 0.03, string.format("%d " .. g_i18n:getText("fluid_unit_long"), self.fuelFillLevel));
    if self.fuelFillLevel < fuelWarn then
        setTextColor(1,1,1,1);
    end;

    setTextBold(false);


    if self.hasRefuelStationInRange and not self.doRefuel and self.fuelFillLevel ~= self.fuelCapacity then
        g_currentMission:addHelpButtonText(g_i18n:getText("Refuel"), InputBinding.REFUEL);
        g_currentMission.hudFuelOverlay:render();
    end;

    if self.showWaterWarning then
        g_currentMission:addWarning(g_i18n:getText("Dont_drive_to_depth_into_the_water"), 0.05, 0.025+0.007);

    end;

    local trailerFillLevel, trailerCapacity = self:getAttachedTrailersFillLevelAndCapacity();
    if trailerFillLevel ~= nil and trailerCapacity ~= nil then
        self:drawGrainLevel(trailerFillLevel, trailerCapacity, 101);
    end;
    if table.getn(self.attachedImplements) > 1 then
        g_currentMission:addHelpButtonText(g_i18n:getText("Change_tools"), InputBinding.SWITCH_IMPLEMENT);
    end;

    if self.isEntered then
	if self.controlhudOverlay ~= nil then
  		if self.panelActive then
			setTextColor(1,1,1,1);
			renderText(0.844, 0.91, 0.021, "0: Info zu");
    			self.controlhudOverlay:render();
		else
			setTextColor(1,1,1,1);
			renderText(0.844, 0.91, 0.021, "0: Info auf");
  		end;
	end;
    end;

end;

function steyer8300:onEnter()
    self.isEntered = true;

    self:startMotor();

    self.camIndex = 1;
    self.cameras[self.camIndex]:onActivate();

    if self.characterNode ~= nil then
        setVisibility(self.characterNode, true);
    end;

    self:onActivateAttachements();
end;

function steyer8300:onLeave()
    self.cameras[self.camIndex]:onDeactivate();

    if self.stopRefuelOnLeave then
        self:stopRefuel();
    end;

    if self.deactivateLightsOnLeave then
        self:setLightsVisibility(false);
    end;

    if self.characterNode ~= nil then
        if self.disableCharacterOnLeave then
            setVisibility(self.characterNode, false);
        else
            setVisibility(self.characterNode, true);
        end;
    end;

    if self.stopMotorOnLeave then
        self:stopMotor();
    else
        Motorized.stopSounds(self);
    end;

    if self.deactivateOnLeave then
        for k,wheel in pairs(self.wheels) do
            setWheelShapeProps(wheel.node, wheel.wheelShape, 0, self.motor.brakeForce, 0);
        end;
        self:onDeactivateAttachements();
    else
        if self.deactivateLightsOnLeave then
            self:onDeactivateAttachementsLights();
        end;
        self:onDeactivateAttachementsSounds();
    end;
    self.isEntered = false;
end;

function steyer8300:drawGrainLevel(level, capacity, warnPercent)

    local percent = 0;
    if capacity > 0 then
        percent = level/capacity*100;
    end;
    setTextBold(true);

    if percent >= warnPercent then
        setTextColor(1.0, 0.0, 0.0, 1.0);
    else
        setTextColor(1.0, 1.0, 1.0, 1.0);
    end;

    renderText(self.hudBasePoxX+0.031, self.hudBasePoxY+0.006, 0.03, string.format("%d(%d%%)", level, percent));
    if percent >= warnPercent then
        setTextColor(1.0, 1.0, 1.0, 1.0);
    end;
    setTextBold(false);
end;

--ai

function steyer8300:startAITractor()
    self:hire();
    if not self.isAITractorActivated then

        self.isAITractorActivated = true;
        self.turnTimer = self.turnTimeoutLong;
        self.turnStage = 0;

        local x,y,z = localDirectionToWorld(self.aiTractorDirectionNode, 0, 0, 1);
        local length = Utils.vector2Length(x,z);
        self.aiTractorDirectionX = x/length;
        self.aiTractorDirectionZ = z/length;

        local x,y,z = getWorldTranslation(self.aiTractorDirectionNode);
        self.aiTractorTargetX = x;
        self.aiTractorTargetZ = z;

        self.numCollidingVehicles = 0;
        if self.aiTrafficCollisionTrigger ~= nil then
            addTrigger(self.aiTrafficCollisionTrigger, "onTrafficCollisionTrigger", self);
        end;

       steyer8300.updateToolsInfo(self);
        for k,implement in pairs(self.attachedImplements) do
            if implement.object.needsLowering and implement.object.aiNeedsLowering then
                local jointDesc = self.attacherJoints[implement.jointDescIndex];
                jointDesc.moveDown = true;
            end;
            implement.object:aiTurnOn();
            steyer8300.addToolTrigger(self, implement.object);
        end;

        self.checkSpeedLimit = false;

        setVisibility(self.aiMotorSound, true);

    end;
end;

function steyer8300:stopAITractor()
    self:dismiss();
    if self.isAITractorActivated then

        self.motor:setSpeedLevel(0, false);
        self.motor.maxRpmOverride = nil;

        WheelsUtil.updateWheelsPhysics(self, 0, self.lastSpeed, 0, false, self.requiredDriveMode);

        if self.aiTrafficCollisionTrigger ~= nil then
            removeTrigger(self.aiTrafficCollisionTrigger);
        end;
        self.isAITractorActivated = false;

        self.checkSpeedLimit = true;

        setVisibility(self.aiMotorSound, false);

        for k,implement in pairs(self.attachedImplements) do
            steyer8300.removeToolTrigger(self, implement.object);
            implement.object:aiTurnOff();
        end;

    end;
end;

function steyer8300.updateAIMovement(self, dt)

    local allowedToDrive = true;

    if self.numCollidingVehicles > 0 then
        allowedToDrive = false;
    end;
    for k,v in pairs(self.numToolsCollidingVehicles) do
        if v > 0 then
            allowedToDrive = false;
            break;
        end;
    end;
        if self.waitForTurnTime > self.time then
            allowedToDrive = false;
        end;
    if not allowedToDrive then
        local lx, lz = 0, 1; --AIVehicleUtil.getDriveDirection(self.aiTractorDirectionNode, self.aiTractorTargetX, y, self.aiTractorTargetZ);
        AIVehicleUtil.driveInDirection(self, dt, 30, 0, 0, 28, false, moveForwards, lx, lz)
        return;
    end;

    local speedLevel = 1;

    local leftMarker = self.aiLeftMarker;
    local rightMarker = self.aiRightMarker;
    local backMarker = self.aiBackMarker;
    local terrainDetailChannel1 = self.aiTerrainDetailChannel1;
    local terrainDetailChannel2 = self.aiTerrainDetailChannel2;

    if leftMarker == nil or rightMarker == nil or backMarker == nil then
        self:stopAITractor();
        return;
    end;

    local newTargetX, newTargetY, newTargetZ;

    local moveForwards = true;
    local updateWheels = true;

    self.turnTimer = self.turnTimer - dt;

    if self.turnTimer < 0 or self.turnStage > 0 then
        if self.turnStage > 1 then
            local x,y,z = getWorldTranslation(self.aiTractorDirectionNode);
            local dirX, dirZ = self.aiTractorDirectionX, self.aiTractorDirectionZ;
            local myDirX, myDirY, myDirZ = localDirectionToWorld(self.aiTractorDirectionNode, 0, 0, 1);

            newTargetX = self.aiTractorTargetX;
            newTargetY = y;
            newTargetZ = self.aiTractorTargetZ;
            if self.turnStage == 2 then
                self.turnStageTimer = self.turnStageTimer - dt;
                if myDirX*dirX + myDirZ*dirZ > 0.2 or self.turnStageTimer < 0 then
                    if self.aiTurnNoBackward then
                        self.turnStage = 4;
                    else
                        self.turnStage = 3;
                        moveForwards = false;
                    end;
                    if self.turnStageTimer < 0 then

                        self.aiTractorTargetBeforeSaveX = self.aiTractorTargetX;
                        self.aiTractorTargetBeforeSaveZ = self.aiTractorTargetZ;

                        newTargetX = self.aiTractorTargetBeforeTurnX;
                        newTargetZ = self.aiTractorTargetBeforeTurnZ;

                        moveForwards = false;
                        self.turnStage = 6;
                        self.turnStageTimer = self.turnStage6Timeout;
                    else
                        self.turnStageTimer = self.turnStage3Timeout;
                    end;
                end;
            elseif self.turnStage == 3 then
                self.turnStageTimer = self.turnStageTimer - dt;
                if myDirX*dirX + myDirZ*dirZ > 0.95 or self.turnStageTimer < 0 then
                    self.turnStage = 4;
                else
                    moveForwards = false;
                end;
            elseif self.turnStage == 4 then
                local dx, dz = x-newTargetX, z-newTargetZ;
                local dot = dx*dirX + dz*dirZ;
                if -dot < self.turnEndDistance then
                    newTargetX = self.aiTractorTargetX + dirX*self.turnTargetMoveBack;
                    newTargetY = y;
                    newTargetZ = self.aiTractorTargetZ + dirZ*self.turnTargetMoveBack;
                    self.turnStage = 5;
                end;
            elseif self.turnStage == 5 then
                local backX, backY, backZ = getWorldTranslation(backMarker);
                local dx, dz = backX-newTargetX, backZ-newTargetZ;
                local dot = dx*dirX + dz*dirZ;
                if -dot < self.turnEndBackDistance+self.aiMarkerLength then
                    self.turnTimer = self.turnTimeoutLong;
                    self.turnStage = 0;
                    for k,implement in pairs(self.attachedImplements) do
                        if implement.object.needsLowering and implement.object.aiNeedsLowering then
                            local jointDesc = self.attacherJoints[implement.jointDescIndex];
                            jointDesc.moveDown = true;
                        end;
                        for k,implement in pairs(self.attachedImplements) do
                            implement.object:aiLower();
                        end;
                    end;
                    self.waitForTurnTime = self.time + self.waitForTurnTimeout;
                    steyer8300.updateInvertLeftRight(self);
                    leftMarker = self.aiLeftMarker;
                    rightMarker = self.aiRightMarker;
                end;
            elseif self.turnStage == 6 then
                self.turnStageTimer = self.turnStageTimer - dt;
                if self.turnStageTimer < 0 then
                    self.turnStageTimer = self.turnStage2Timeout;
                    self.turnStage = 2;

                    newTargetX = self.aiTractorTargetBeforeSaveX;
                    newTargetZ = self.aiTractorTargetBeforeSaveZ;
                else
                    local x,y,z = getWorldTranslation(self.aiTractorDirectionNode);
                    local dirX, dirZ = -self.aiTractorDirectionX, -self.aiTractorDirectionZ;
                    local targetX, targetZ = self.aiTractorTargetX, self.aiTractorTargetZ;
                    local dx, dz = x-targetX, z-targetZ;
                    local dot = dx*dirX + dz*dirZ;

                    local projTargetX = targetX +dirX*dot;
                    local projTargetZ = targetZ +dirZ*dot;

                    newTargetX = projTargetX-dirX*self.aiTractorLookAheadDistance;
                    newTargetZ = projTargetZ-dirZ*self.aiTractorLookAheadDistance;
                    moveForwards = false;
                end;
            end;
        elseif self.turnStage == 1 then
            steyer8300.updateInvertLeftRight(self);
            leftMarker = self.aiLeftMarker;
            rightMarker = self.aiRightMarker;

            local x,y,z = getWorldTranslation(self.aiTractorDirectionNode);
            local dirX, dirZ = self.aiTractorDirectionX, self.aiTractorDirectionZ;
            local sideX, sideZ = -dirZ, dirX;
            local lX,  lY,  lZ = getWorldTranslation(leftMarker);
            local rX,  rY,  rZ = getWorldTranslation(rightMarker);

            local markerWidth = Utils.vector2Length(lX-rX, lZ-rZ);
            local turnLeft = true;

            local lWidthX = lX + dirX * self.sideWatchDirOffset;
            local lWidthZ = lZ + dirZ * self.sideWatchDirOffset;
            local lStartX = lWidthX - sideX*0.7*markerWidth;
            local lStartZ = lWidthZ - sideZ*0.7*markerWidth;
            local lHeightX = lStartX + dirX*self.sideWatchDirSize;
            local lHeightZ = lStartZ + dirZ*self.sideWatchDirSize;

            local rWidthX = rX + dirX * self.sideWatchDirOffset;
            local rWidthZ = rZ + dirZ * self.sideWatchDirOffset;
            local rStartX = rWidthX + sideX*0.7*markerWidth;
            local rStartZ = rWidthZ + sideZ*0.7*markerWidth;
            local rHeightX = rStartX + dirX*self.sideWatchDirSize;
            local rHeightZ = rStartZ + dirZ*self.sideWatchDirSize;

            local leftArea = 0;
            if terrainDetailChannel1 >= 0 then
                local id = g_currentMission.terrainDetailId;
                local x,z, widthX,widthZ, heightX,heightZ = Utils.getXZWidthAndHeight(id, lStartX, lStartZ, lWidthX, lWidthZ, lHeightX, lHeightZ);
                leftArea = getDensityParallelogram(id, x, z, widthX, widthZ, heightX, heightZ, terrainDetailChannel1, 1);
                if leftArea == 0 and terrainDetailChannel2 >= 0 then
                    leftArea = getDensityParallelogram(id, x, z, widthX, widthZ, heightX, heightZ, terrainDetailChannel2, 1);
                end;
            end;

            local rightArea = 0;
            if terrainDetailChannel1 >= 0 then
                local id = g_currentMission.terrainDetailId;
                local x,z, widthX,widthZ, heightX,heightZ = Utils.getXZWidthAndHeight(id, rStartX, rStartZ, rWidthX, rWidthZ, rHeightX, rHeightZ);
                rightArea = getDensityParallelogram(id, x, z, widthX, widthZ, heightX, heightZ, terrainDetailChannel1, 1);
                if rightArea == 0 and terrainDetailChannel2 >= 0 then
                    rightArea = getDensityParallelogram(id, x, z, widthX, widthZ, heightX, heightZ, terrainDetailChannel2, 1);
                end;
            end;
            if leftArea > 0 or rightArea > 0 then
                if leftArea > rightArea then
                    turnLeft = true;
                else
                    turnLeft = false;
                end
            else
                self:stopAITractor();
                return;
            end;

            local targetX, targetZ = self.aiTractorTargetX, self.aiTractorTargetZ;
            markerWidth = markerWidth*self.aiTurnWidthScale;

            local invertsMarker = steyer8300.invertsMarkerOnTurn(self, turnLeft);
            if not invertsMarker then
                local mx = (lX+rX)/2;
                local mz = (lZ+rZ)/2;
                local markerSideOffset, t1, t2 = worldToLocal(self.aiTractorDirectionNode, mx, (lY+rY)/2, mz);
                markerWidth = markerWidth + markerSideOffset;
            end;


            local backX, backY, backZ = getWorldTranslation(backMarker);
            local projTargetX, projTargetZ = Utils.projectOnLine(backX, backZ, targetX, targetZ, dirX, dirZ)

            x = projTargetX;
            z = projTargetZ;

            if turnLeft then
                newTargetX = x-sideX*markerWidth;
                newTargetY = y;
                newTargetZ = z-sideZ*markerWidth;
                for k,implement in pairs(self.attachedImplements) do
                    implement.object:aiRotateLeft();
                end;
            else
                newTargetX = x+sideX*markerWidth;
                newTargetY = y;
                newTargetZ = z+sideZ*markerWidth;
                for k,implement in pairs(self.attachedImplements) do
                    implement.object:aiRotateRight();
                end;
            end;

            self.aiTurnNoBackward = markerWidth >= 2*self.aiTractorTurnRadius;

            self.aiTractorTargetBeforeTurnX = self.aiTractorTargetX;
            self.aiTractorTargetBeforeTurnZ = self.aiTractorTargetZ;

            self.aiTractorDirectionX = -dirX;
            self.aiTractorDirectionZ = -dirZ;

            self.turnStage = 2;
            self.turnStageTimer = self.turnStage2Timeout;
        else
            self.turnStage = 1;

            self.waitForTurnTime = self.time + self.waitForTurnTimeout;
            for k,implement in pairs(self.attachedImplements) do
                if implement.object.needsLowering and implement.object.aiNeedsLowering then
                    local jointDesc = self.attacherJoints[implement.jointDescIndex];
                    jointDesc.moveDown = false;
                end;
                for k,implement in pairs(self.attachedImplements) do
                    implement.object:aiRaise();
                end;
            end;
            updateWheels = false;
        end;
    else
        local dirX, dirZ = self.aiTractorDirectionX, self.aiTractorDirectionZ;
        local lX,  lY,  lZ = getWorldTranslation(leftMarker);
        local rX,  rY,  rZ = getWorldTranslation(rightMarker);
        self.lastFrontMarkerDistance = self.lastSpeed*self.turnTimeout;
        local scaledDistance = self.lastFrontMarkerDistance*self.frontMarkerDistanceScale
        lX = lX + dirX*scaledDistance;
        lZ = lZ + dirZ*scaledDistance;

        rX = rX + dirX*scaledDistance;
        rZ = rZ + dirZ*scaledDistance;

        local heightX = lX + dirX*2;
        local heightZ = lZ + dirZ*2;

        local area = 0;
        if terrainDetailChannel1 >= 0 then
            local id = g_currentMission.terrainDetailId;
            local x,z, widthX,widthZ, heightX,heightZ = Utils.getXZWidthAndHeight(id, lX, lZ, rX, rZ, heightX, heightZ);
            area = getDensityParallelogram(id, x, z, widthX, widthZ, heightX, heightZ, terrainDetailChannel1, 1);
            if area == 0 and terrainDetailChannel2 >= 0 then
                area = getDensityParallelogram(id, x, z, widthX, widthZ, heightX, heightZ, terrainDetailChannel2, 1);
            end;
        end;

        if area >= 1 then
            self.turnTimer = self.turnTimeout;
        end;


        local x,y,z = getWorldTranslation(self.aiTractorDirectionNode);
        local dirX, dirZ = self.aiTractorDirectionX, self.aiTractorDirectionZ;
        local targetX, targetZ = self.aiTractorTargetX, self.aiTractorTargetZ;
        local dx, dz = x-targetX, z-targetZ;
        local dot = dx*dirX + dz*dirZ;

        local projTargetX = targetX +dirX*dot;
        local projTargetZ = targetZ +dirZ*dot;

        newTargetX = projTargetX+self.aiTractorDirectionX*self.aiTractorLookAheadDistance;
        newTargetY = y;
        newTargetZ = projTargetZ+self.aiTractorDirectionZ*self.aiTractorLookAheadDistance;
    end;

    if updateWheels then
        local lx, lz = AIVehicleUtil.getDriveDirection(self.aiTractorDirectionNode, newTargetX, newTargetY, newTargetZ);

        if self.turnStage == 3 and math.abs(lx) < 0.1 then
            self.turnStage = 4;
            moveForwards = true;
        end;

        AIVehicleUtil.driveInDirection(self, dt, 25, 0.5, 0.5, 20, true, moveForwards, lx, lz, speedLevel, 0.9);

        local maxlx = 0.7071067; --math.sin(maxAngle);
        local colDirX = lx;
        local colDirZ = lz;

        if colDirX > maxlx then
            colDirX = maxlx;
            colDirZ = 0.7071067; --math.cos(maxAngle);
        elseif colDirX < -maxlx then
            colDirX = -maxlx;
            colDirZ = 0.7071067; --math.cos(maxAngle);
        end;

        if self.aiTrafficCollisionTrigger ~= nil then
            AIVehicleUtil.setCollisionDirection(self.aiTractorDirectionNode, self.aiTrafficCollisionTrigger, colDirX, colDirZ);
        end;
        for k,v in pairs(self.numToolsCollidingVehicles) do
            AIVehicleUtil.setCollisionDirection(self.aiTractorDirectionNode, k, colDirX, colDirZ);
        end;
    end;

    if newTargetX ~= nil and newTargetZ ~= nil then
        self.aiTractorTargetX = newTargetX;
        self.aiTractorTargetZ = newTargetZ;
    end;
end;

function steyer8300.switchToDirection(self, myDirX, myDirZ)
    self.aiTractorDirectionX = myDirX;
    self.aiTractorDirectionZ = myDirZ;
end;

function steyer8300.addToolTrigger(self, tool)
    if tool.aiTrafficCollisionTrigger ~= nil then
        addTrigger(tool.aiTrafficCollisionTrigger, "onToolTrafficCollisionTrigger", self);
        self.numToolsCollidingVehicles[tool.aiTrafficCollisionTrigger] = 0;
    end;
    for k,v in pairs(tool.components) do
        self.trafficCollisionIgnoreList[v.node] = true;
    end;
end;

function steyer8300.removeToolTrigger(self, tool)
    if tool.aiTrafficCollisionTrigger ~= nil then
        removeTrigger(tool.aiTrafficCollisionTrigger);
        self.numToolsCollidingVehicles[tool.aiTrafficCollisionTrigger] = nil;
    end;
    for k,v in pairs(tool.components) do
        self.trafficCollisionIgnoreList[v.node] = nil;
    end;
end;


function steyer8300:attachImplement(implement)
    local object = implement.object;
    if self.isAITractorActivated then
        steyer8300.addToolTrigger(self, object);
    end;
    self.aiToolsDirty = true;
end;

function steyer8300:detachImplement(implementIndex)
    local object = self.attachedImplements[implementIndex].object;
    if self.isAITractorActivated then
        steyer8300.removeToolTrigger(self, object);
    end;
    self.aiToolsDirty = true;
end;

function steyer8300:onTrafficCollisionTrigger(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
    if onEnter or onLeave then
        if otherId == Player.rootNode then
            if onEnter then
                self.numCollidingVehicles = self.numCollidingVehicles+1;
            elseif onLeave then
                self.numCollidingVehicles = math.max(self.numCollidingVehicles-1, 0);
            end;
        else
            local vehicle = g_currentMission.nodeToVehicle[otherId];
            if vehicle ~= nil and self.trafficCollisionIgnoreList[otherId] == nil then
                if onEnter then
                    self.numCollidingVehicles = self.numCollidingVehicles+1;
                elseif onLeave then
                    self.numCollidingVehicles = math.max(self.numCollidingVehicles-1, 0);
                end;
            end;
        end;
    end;
end;

function steyer8300:onToolTrafficCollisionTrigger(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
    if onEnter or onLeave then
        if otherId == Player.rootNode then
            if onEnter then
                self.numToolsCollidingVehicles[triggerId] = self.numToolsCollidingVehicles[triggerId]+1;
            elseif onLeave then
                self.numToolsCollidingVehicles[triggerId] = math.max(self.numToolsCollidingVehicles[triggerId]-1, 0);
            end;
        else
            local vehicle = g_currentMission.nodeToVehicle[otherId];
            if vehicle ~= nil and self.trafficCollisionIgnoreList[otherId] == nil then
                if onEnter then
                    self.numToolsCollidingVehicles[triggerId] = self.numToolsCollidingVehicles[triggerId]+1;
                elseif onLeave then
                    self.numToolsCollidingVehicles[triggerId] = math.max(self.numToolsCollidingVehicles[triggerId]-1, 0);
                end;
            end;
        end;
    end;
end;

function steyer8300.updateToolsInfo(self)
    local leftMarker = nil;
    local rightMarker = nil;
    local backMarker = nil;
    local terrainDetailChannel1 = -1;
    local terrainDetailChannel2 = -1;
    for k,implement in pairs(self.attachedImplements) do
        local object = implement.object;
        if object.aiLeftMarker ~= nil then
            if leftMarker == nil then
                leftMarker = object.aiLeftMarker;
            end;
        end;
        if object.aiRightMarker ~= nil then
            if rightMarker == nil then
                rightMarker = object.aiRightMarker;
            end;
        end;
        if object.aiBackMarker ~= nil then
            if backMarker == nil then
                backMarker = object.aiBackMarker;
            end;
        end;
        if terrainDetailChannel1 < 0 then
            if object.aiTerrainDetailChannel1 >= 0 then
                terrainDetailChannel1 = object.aiTerrainDetailChannel1;
                if object.aiTerrainDetailChannel2 >= 0 then
                    terrainDetailChannel2 = object.aiTerrainDetailChannel2;
                end;
            end;
        end;
    end;
    self.aiLeftMarker = leftMarker;
    self.aiRightMarker = rightMarker;
    self.aiBackMarker = backMarker;
    self.aiTerrainDetailChannel1 = terrainDetailChannel1;
    self.aiTerrainDetailChannel2 = terrainDetailChannel2;
    steyer8300.updateInvertLeftRight(self);

    if leftMarker ~= nil and rightMarker ~= nil and backMarker ~= nil then
        local lX, lY, lZ = worldToLocal(self.aiTractorDirectionNode, getWorldTranslation(leftMarker));
        local rX, rY, rZ = worldToLocal(self.aiTractorDirectionNode, getWorldTranslation(rightMarker));
        local bX, bY, bZ = worldToLocal(self.aiTractorDirectionNode, getWorldTranslation(backMarker));
        local z = (lZ + rZ)/2;
        self.aiMarkerLength = z-bZ;
    end;

    self.aiToolsDirty = false;
end;

function steyer8300.updateInvertLeftRight(self)
    if self.aiLeftMarker ~= nil and self.aiRightMarker ~= nil then
        local lX, lY, lZ = worldToLocal(self.aiTractorDirectionNode, getWorldTranslation(self.aiLeftMarker));
        local rX, rY, rZ = worldToLocal(self.aiTractorDirectionNode, getWorldTranslation(self.aiRightMarker));
        if rX > lX then
            local t = self.aiLeftMarker;
            self.aiLeftMarker = self.aiRightMarker;
            self.aiRightMarker = t;
        end;
    end;
end;

function steyer8300.invertsMarkerOnTurn(self, turnLeft)
    local res = false;
    for k,implement in pairs(self.attachedImplements) do
        for k1, spec in pairs(implement.object.specializations) do
            if spec.aiInvertsMarkerOnTurn then
                res = res or spec.aiInvertsMarkerOnTurn(implement.object, turnLeft);
            end;
        end;
    end;
    return res;
end;
